/*------------------------------------------------------------------------
**------------------------------------------------------------------------
**
** @MODULE	%Z%%M% %I% %G% %U%
**
** @COPYRIGHT	(c)1998 - Global Atmospherics, Inc.
**
** @INDEX	
**
** @AUDIENCE	
**
** @DESCRIPTION
**
**	
**
** @LOG
**
**	03/29/98 21:59:10	@author(dhagberg)
**
**		Created initial version.
**
**------------------------------------------------------------------------
**----------------------------------------------------------------------*/

	/*--------------------------------------------------------------  
	  #include
	  --------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioccom.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sundev/syncstat.h>
#include <netx25/x25_pk.h>
#include <netx25/x25_ctl.h>
#include <netx25/x25_ioctl.h>

#include <tcl.h>

	/*--------------------------------------------------------------  
	  #define
	  --------------------------------------------------------------*/

/*
 * These bits may be ORed together into the "flags" field of a X25State
 * structure.
 */

#define X25_ASYNC_SOCKET	(1<<0)	/* Asynchronous socket. */
#define X25_ASYNC_CONNECT	(1<<1)	/* Async connect in progress. */

/*
 * The following defines the maximum length of the listen queue. This is
 * the number of outstanding yet-to-be-serviced requests for a connection
 * on a server socket, more than this number of outstanding requests and
 * the connection request will fail.
 */

#ifndef	SOMAXCONN
#define SOMAXCONN	100
#endif

#if	(SOMAXCONN < 100)
#undef	SOMAXCONN
#define	SOMAXCONN	100
#endif

/*
 * The following defines how much buffer space the kernel should maintain
 * for a socket.
 */

#define SOCKET_BUFSIZE	4096

	/*--------------------------------------------------------------  
	  struct/union/enum/typedef
	  --------------------------------------------------------------*/

/*
 * This callback function type is responsible for accepting X25 connections
 */

typedef void (X25AcceptProc) _ANSI_ARGS_((ClientData callbackData,
        Tcl_Channel chan, char *remote_addr, int link, int lcn));

/*
 * This structure describes per-instance state of an X.25 based channel.
 */

typedef struct X25State {
    Tcl_Channel channel;	/* Channel associated with this file. */
    int fd;			/* The socket itself. */
    int flags;			/* ORed combination of the bitfields
				 * X25_* defined above. */
    X25AcceptProc *acceptProc;	/* Proc to call on accept. */
    ClientData acceptProcData;	/* The data for the accept proc. */

    int numRemoteAddrs;		/* Number of remote addresses in list below */
    CONN_ADR *remoteAddrList;	/* List of remote addresses for Server to
				 * accept.  Accepts everything if NULL or if
				 * numRemoteAddrs is 0 */
} X25State;

/*
 * Callback structure for accept callback in a TCP server.
 */

typedef struct X25AcceptCallback {
    char *script;			/* Script to invoke. */
    Tcl_Interp *interp;			/* Interpreter in which to run it. */
} X25AcceptCallback;

	/*--------------------------------------------------------------
	  prototypes
	  --------------------------------------------------------------*/

	/* Exported Functions to handle Tcl commands for X.25		*/
int X25_Init(Tcl_Interp *interp);
int X25_SocketObjCmd(ClientData notUsed, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[]);
int X25_HostListObjCmd(ClientData notused, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[]);

	/* Internal Functions to clean up and handle X.25 server sockets*/
static void X25AcceptCallbackProc(ClientData callbackData, Tcl_Channel chan,
        char *remote_addr, int link, int lcn); /*OK*/
static void X25RegisterServerInterpCleanup(Tcl_Interp *interp,
        X25AcceptCallback *acceptCallbackPtr);
static void X25AcceptCallbacksDeleteProc(ClientData clientData,
        Tcl_Interp *interp);
static void X25ServerCloseProc(ClientData callbackData);
static void X25UnregisterServerInterpCleanupProc(Tcl_Interp *interp,
        X25AcceptCallback *acceptCallbackPtr);
static void X25Accept(ClientData data, int mask);

	/* Internal Functions to handle file operations 		*/
static int X25BlockModeProc(ClientData instanceData, int mode);
static int X25WaitForConnect(X25State *statePtr, int *errorCodePtr);
static int X25InputProc(ClientData instanceData, char *buf, int bufSize,
        int *errorCodePtr);
static int X25OutputProc(ClientData instanceData, char *buf, int toWrite,
        int *errorCodePtr);
static int X25CloseProc(ClientData instanceData, Tcl_Interp *interp);
static int X25GetOptionProc(ClientData instanceData, Tcl_Interp *interp,
        char *optionName, Tcl_DString *dsPtr);
static void X25WatchProc(ClientData instanceData, int mask);
static int X25GetHandleProc(ClientData instanceData, int direction,
        ClientData *handlePtr);

	/* Internal Functions for encoding/decoding/comparing addresses	*/
static int DecodeBCDAddr(char* addr, int addrlen, const unsigned char* host,
        int hostlen);
static int EncodeBCDAddr(unsigned char *dest, unsigned char *destlen,
        const char* addr);
static int CompareBCDAddr(unsigned char len1, unsigned char *addr1,
        unsigned char len2, unsigned char *addr2);
static int X25AddrListToBCDArray(Tcl_Interp *inter, CONN_ADR *bcdArrPtr[],
        int *bcdArrLenPtr, Tcl_Obj *addrList);
static int X25BCDArrayToAddrList(Tcl_Interp *interp, Tcl_Obj *addrList,
	CONN_ADR bcdArr[], int bcdArrLen);

	/* Internal function for creating X.25 server & client sockets	*/
static X25State *X25CreateSocket(Tcl_Interp *interp, Tcl_Obj *hostList,
        int link, char *localaddr, int server, int async);

	/* Exported functions for creating X.25 client/server sockets	*/
Tcl_Channel X25_OpenClient(Tcl_Interp *interp, char *address, int link,
        char *localaddr, int async);
Tcl_Channel X25_MakeClientChannel(ClientData sock);
Tcl_Channel X25_OpenServer(Tcl_Interp *interp, Tcl_Obj *hostList, int link,
        char* localaddr, X25AcceptProc *acceptProc, ClientData acceptProcData);

	/* internal functions from Tcl library				*/
EXTERN int TclUnixWaitForFile(int fd, int mask, int timeout);
EXTERN int TclSockMinimumBuffers(int sock, int size);
EXTERN int TclFormatInt(char *buffer, long n);
EXTERN void panic TCL_VARARGS(char *,format);


	/*--------------------------------------------------------------
	  local/global data
	  --------------------------------------------------------------*/

/*
 * This structure describes the channel type structure for X25 socket
 * based IO:
 */

static Tcl_ChannelType x25ChannelType = {
    "x25",				/* Type name. */
    X25BlockModeProc,			/* Set blocking/nonblocking mode.*/
    X25CloseProc,			/* Close proc. */
    X25InputProc,			/* Input proc. */
    X25OutputProc,			/* Output proc. */
    NULL,				/* Seek proc. */
    NULL,				/* Set option proc. */
    X25GetOptionProc,			/* Get option proc. */
    X25WatchProc,			/* Initialize notifier. */
    X25GetHandleProc,			/* Get OS handles out of channel. */
};

	/*--------------------------------------------------------------
	  code
	  --------------------------------------------------------------*/

/*
 *----------------------------------------------------------------------
 *
 * X25_Init --
 *
 *	This procedure initializes an interpreter with X.25-related
 *	commands.
 *
 * Results:
 *	Returns TCL_OK on success, TCL_ERROR otherwise.
 *
 * Side effects:
 *	Commands added to Tcl interpreter.
 *
 *----------------------------------------------------------------------
 */

int
X25_Init(Tcl_Interp *interp)
{
    Tcl_PkgProvide(interp,"x25","0.1.0");
    Tcl_CreateObjCommand(interp, "x25::hostlist", X25_HostListObjCmd,
            (ClientData) NULL, (Tcl_CmdDeleteProc*) NULL);
    Tcl_CreateObjCommand(interp, "x25::socket",   X25_SocketObjCmd,
            (ClientData) NULL, (Tcl_CmdDeleteProc*) NULL);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * X25_HostListObjCmd --
 *
 *	This procedure is invoked to process the "x25::hostlist" Tcl command,
 *	which is used to modify/view the list of accepted remote addresses
 *	on an open X.25 SERVER channel.  Has no effect on a client channel.
 *
 * Results:
 *	When called with only a channel argument, returns the current list
 *	of accepted addresses.  When called with a channel + list of
 *	addresses, returns nothing.
 *
 * Side effects:
 *
 *	None when calld with only a channel argument.  When called
 *	with a channel argument and a second argument with a list of
 *	X.25 address strings, it will modify the current list of
 *	accepted addresses (if all address strings are valid X.25
 *	addresses)
 *
 *----------------------------------------------------------------------
 */

int
X25_HostListObjCmd(ClientData notUsed,	/* Not used.			*/
        Tcl_Interp *interp,		/* Current Interpreter.		*/
        int objc,			/* number of arguments		*/
        Tcl_Obj *CONST objv[])		/* argument object array	*/
{
    int i, numAddrs;
    char buf[20];
    Tcl_Channel chan;
    X25State *state;
    Tcl_Obj *addrList = NULL;

    if( (objc < 2) || (objc > 3) ) {
	Tcl_WrongNumArgs(interp, 1, objv,"x25chan ?list?");
	return TCL_ERROR;
    }
    if( (chan = Tcl_GetChannel(interp, Tcl_GetStringFromObj(objv[1],NULL),
            NULL)) == NULL ) {
	return TCL_ERROR;
    }
    if( Tcl_GetChannelType(chan) != &x25ChannelType ) {
        Tcl_AppendResult(interp, "not a valid X.25 channel", (char *) NULL);
	return TCL_ERROR;
    }
    if( (state = (X25State *) Tcl_GetChannelInstanceData(chan)) == NULL ) {
        return TCL_ERROR;
    }

    if( (addrList = Tcl_NewListObj(0,NULL)) == NULL ) {
        return TCL_ERROR;
    }
    Tcl_IncrRefCount(addrList);
    if( objc == 2 ) {
        if( X25BCDArrayToAddrList(interp, addrList, state->remoteAddrList,
                state->numRemoteAddrs) != TCL_OK ) {
	    return TCL_ERROR;
	}
	Tcl_SetObjResult(interp,addrList);
    } else {
        if( Tcl_ListObjAppendList(interp,addrList,objv[2]) != TCL_OK ) {
            Tcl_DecrRefCount(addrList);
            return TCL_ERROR;
	}
	if( X25AddrListToBCDArray(interp,&(state->remoteAddrList),
                &(state->numRemoteAddrs), addrList) != TCL_OK ) {
            Tcl_DecrRefCount(addrList);
            return TCL_ERROR;
	}
    }
    Tcl_DecrRefCount(addrList);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * X25_SocketObjCmd --
 *
 *	This procedure is invoked to process the "x25::socket" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Creates a socket based channel.
 *
 *----------------------------------------------------------------------
 */

int
X25_SocketObjCmd(ClientData notUsed,	/* Not used.			*/
        Tcl_Interp *interp,		/* Current Interpreter.		*/
        int objc,			/* number of arguments		*/
        Tcl_Obj *CONST objv[])		/* argument object array	*/
{
    int a, len;
    char *arg, *copyScript;
    Tcl_Obj *lastHostObj	= NULL;
    char *localaddr		= NULL;
    int server			= 0;
    char *script		= NULL;
    int async			= 0;
    int link			= -1;
    int numHosts		= 0;
    Tcl_Obj *hostList		= NULL;
    Tcl_Obj *tmpStr 		= NULL;
    Tcl_Channel chan;
    X25AcceptCallback *acceptCallbackPtr;

    for (a = 1; a < objc; a++) {
        arg = Tcl_GetStringFromObj(objv[a], &len);
	if (arg[0] == '-') {
	    if (strncmp(arg, "-server", len) == 0) {
                if (async == 1) {
                    Tcl_AppendResult(interp,
                            "cannot set -async option for X.25 server sockets",
                            (char *) NULL);
                    return TCL_ERROR;
                }
		server = 1;
		a++;
		if (a >= objc) {
		    Tcl_AppendResult(interp,
			    "no argument given for -server option",
                            (char *) NULL);
		    return TCL_ERROR;
		}
                script = Tcl_GetStringFromObj(objv[a],NULL);
            } else if (strncmp(arg, "-localaddr", len) == 0) {
		a++;
                if (a >= objc) {
		    Tcl_AppendResult(interp,
			    "no argument given for -localaddr option",
                            (char *) NULL);
		    return TCL_ERROR;
		}
		localaddr = Tcl_GetStringFromObj(objv[a],NULL);
	    } else if (strncmp(arg, "-link", len) == 0) {
	        a++;
		if (a >= objc) {
		    Tcl_AppendResult(interp,
                            "no argument given for -link option",
			    (char *) NULL);
		    return TCL_ERROR;
		}
                if( Tcl_GetIntFromObj(interp, objv[a], &link) != TCL_OK ) {
                    Tcl_AppendResult(interp,
	                    "invalid number passed for -link option",
			    (char *) NULL);
		    return TCL_ERROR;
		}
            } else if (strncmp(arg, "-async", len) == 0) {
                if (server == 1) {
                    Tcl_AppendResult(interp,
                            "cannot set -async option for server sockets",
                            (char *) NULL);
                    return TCL_ERROR;
                }
                async = 1;
	    } else {
		Tcl_AppendResult(interp, "bad option \"", arg,
                        "\", must be -async, -localaddr, -link, or -server",
                        (char *) NULL);
		return TCL_ERROR;
	    }
	} else {
	    break;
	}
    }

    if( (hostList = Tcl_NewListObj(0,NULL)) == NULL ) {
        return TCL_ERROR;
    }
    Tcl_IncrRefCount(hostList);
    for( ; a < objc; a++ ) {
        if( Tcl_ListObjAppendElement(interp,hostList,objv[a]) != TCL_OK ) {
	    Tcl_DecrRefCount(hostList);
	    return TCL_ERROR;
	}
	numHosts++;
    }

    if( a != objc ) {
wrongNumArgs:
        arg = Tcl_GetStringFromObj(objv[0],NULL);
	Tcl_AppendResult(interp, "wrong # args: should be either:\n",
		arg," ?-localaddr addr? ?-link linknum? ?-async? remoteaddr\n",
                arg," -server command ?-localaddr addr? ?-link linknum? "
                "remoteaddr ?remoteaddr...?",
                (char *) NULL);
	if( hostList != NULL ) {
            Tcl_DecrRefCount(hostList);
	}
        return TCL_ERROR;
    }

    if (server) {
        acceptCallbackPtr = (X25AcceptCallback *) ckalloc((unsigned)
                sizeof(X25AcceptCallback));
	copyScript = ckalloc((unsigned) strlen(script) + 1);
	strcpy(copyScript, script);
        acceptCallbackPtr->script = copyScript;
        acceptCallbackPtr->interp = interp;
        chan = X25_OpenServer(interp, hostList, link, localaddr,
                X25AcceptCallbackProc, (ClientData) acceptCallbackPtr);
	Tcl_DecrRefCount(hostList);
        if (chan == (Tcl_Channel) NULL) {
	    ckfree(copyScript);
            ckfree((char *) acceptCallbackPtr);
            return TCL_ERROR;
        }

        /*
         * Register with the interpreter to let us know when the
         * interpreter is deleted (by having the callback set the
         * acceptCallbackPtr->interp field to NULL). This is to
         * avoid trying to eval the script in a deleted interpreter.
         */

        X25RegisterServerInterpCleanup(interp, acceptCallbackPtr);
        
        /*
         * Register a close callback. This callback will inform the
         * interpreter (if it still exists) that this channel does not
         * need to be informed when the interpreter is deleted.
         */
        
        Tcl_CreateCloseHandler(chan, X25ServerCloseProc,
                (ClientData) acceptCallbackPtr);
    } else {
        if( numHosts != 1 ) {
	    goto wrongNumArgs;
	}
	if( Tcl_ListObjIndex(interp, hostList, 0, &lastHostObj) != TCL_OK ) {
	    goto wrongNumArgs;
	}
        chan = X25_OpenClient(interp, Tcl_GetStringFromObj(lastHostObj,NULL),
                link, localaddr, async);
	Tcl_DecrRefCount(hostList);
        if (chan == (Tcl_Channel) NULL) {
            return TCL_ERROR;
        }
    }
    Tcl_RegisterChannel(interp, chan);
    Tcl_AppendResult(interp, Tcl_GetChannelName(chan), (char *) NULL);
    
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * X25AcceptCallbackProc --
 *
 *	This callback is invoked by the TCP channel driver when it
 *	accepts a new connection from a client on a server socket.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Whatever the script does.
 *
 *----------------------------------------------------------------------
 */

static void
X25AcceptCallbackProc(
        ClientData callbackData,	/* The data stored when the callback
                                         * was created in the call to
                                         * Tcl_OpenTcpServer. */
        Tcl_Channel chan,		/* Channel for the newly accepted
                                         * connection. */
        char *address,			/* Remote X.25 Address accepted */
        int link,			/* Link number for accepted socket */
        int lcn)			/* Logical channel for accepted sock */
{
    X25AcceptCallback *acceptCallbackPtr;
    Tcl_Interp *interp;
    char *script;
    char linkBuf[10];
    char lcnBuf[10];
    int result;

    acceptCallbackPtr = (X25AcceptCallback *) callbackData;

fprintf(stderr,"DBG: X25AcceptCallbackProc started\n");

    /*
     * Check if the callback is still valid; the interpreter may have gone
     * away, this is signalled by setting the interp field of the callback
     * data to NULL.
     */
    
    if (acceptCallbackPtr->interp != (Tcl_Interp *) NULL) {

        script = acceptCallbackPtr->script;
        interp = acceptCallbackPtr->interp;
        
        Tcl_Preserve((ClientData) script);
        Tcl_Preserve((ClientData) interp);

fprintf(stderr,"DBG: X25AcceptCallbackProc running script: %s\n",script);

	TclFormatInt(linkBuf, link);
	TclFormatInt(lcnBuf,  lcn);
        Tcl_RegisterChannel(interp, chan);
        result = Tcl_VarEval(interp, script, " ", Tcl_GetChannelName(chan), 
                " ", address, " ", linkBuf, " ", lcnBuf, (char *) NULL);
        if (result != TCL_OK) {
fprintf(stderr,"DBG: X25AcceptCallbackProc error running script.\n");
            Tcl_BackgroundError(interp);
	    Tcl_UnregisterChannel(interp, chan);
        }
        Tcl_Release((ClientData) interp);
        Tcl_Release((ClientData) script);
    } else {

fprintf(stderr,"DBG: X25AcceptCallbackProc: no interpreter!\n");

        /*
         * The interpreter has been deleted, so there is no useful
         * way to utilize the client socket - just close it.
         */

        Tcl_Close((Tcl_Interp *) NULL, chan);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * X25RegisterServerInterpCleanup --
 *
 *	Registers an accept callback record to have its interp
 *	field set to NULL when the interpreter is deleted.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When, in the future, the interpreter is deleted, the interp
 *	field of the accept callback data structure will be set to
 *	NULL. This will prevent attempts to eval the accept script
 *	in a deleted interpreter.
 *
 *----------------------------------------------------------------------
 */

static void
X25RegisterServerInterpCleanup(
        Tcl_Interp *interp,	/* Interpreter for which we want to be
                                 * informed of deletion. */
        X25AcceptCallback *acceptCallbackPtr)
    				/* The accept callback record whose
                                 * interp field we want set to NULL when
                                 * the interpreter is deleted. */
{
    Tcl_HashTable *hTblPtr;	/* Hash table for accept callback
                                 * records to smash when the interpreter
                                 * will be deleted. */
    Tcl_HashEntry *hPtr;	/* Entry for this record. */
    int new;			/* Is the entry new? */

    hTblPtr = (Tcl_HashTable *) Tcl_GetAssocData(interp, "x25AcceptCallbacks",
            NULL);
    if (hTblPtr == (Tcl_HashTable *) NULL) {
        hTblPtr = (Tcl_HashTable *) ckalloc((unsigned) sizeof(Tcl_HashTable));
        Tcl_InitHashTable(hTblPtr, TCL_ONE_WORD_KEYS);
        (void) Tcl_SetAssocData(interp, "x25AcceptCallbacks",
                X25AcceptCallbacksDeleteProc, (ClientData) hTblPtr);
    }
    hPtr = Tcl_CreateHashEntry(hTblPtr, (char *) acceptCallbackPtr, &new);
    if (!new) {
        panic("X25RegisterTcpServerCleanup: damaged accept record table");
    }
    Tcl_SetHashValue(hPtr, (ClientData) acceptCallbackPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * X25AcceptCallbacksDeleteProc --
 *
 *	Assocdata cleanup routine called when an interpreter is being
 *	deleted to set the interp field of all the accept callback records
 *	registered with	the interpreter to NULL. This will prevent the
 *	interpreter from being used in the future to eval accept scripts.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Deallocates memory and sets the interp field of all the accept
 *	callback records to NULL to prevent this interpreter from being
 *	used subsequently to eval accept scripts.
 *
 *----------------------------------------------------------------------
 */

static void
X25AcceptCallbacksDeleteProc(
    ClientData clientData,	/* Data which was passed when the assocdata
                                 * was registered. */
    Tcl_Interp *interp)		/* Interpreter being deleted - not used. */
{
    Tcl_HashTable *hTblPtr;
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch hSearch;
    X25AcceptCallback *acceptCallbackPtr;

    hTblPtr = (Tcl_HashTable *) clientData;
    for (hPtr = Tcl_FirstHashEntry(hTblPtr, &hSearch);
             hPtr != (Tcl_HashEntry *) NULL;
             hPtr = Tcl_NextHashEntry(&hSearch)) {
        acceptCallbackPtr = (X25AcceptCallback *) Tcl_GetHashValue(hPtr);
        acceptCallbackPtr->interp = (Tcl_Interp *) NULL;
    }
    Tcl_DeleteHashTable(hTblPtr);
    ckfree((char *) hTblPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * X25ServerCloseProc --
 *
 *	This callback is called when the X.25 server channel for which it
 *	was registered is being closed. It informs the interpreter in
 *	which the accept script is evaluated (if that interpreter still
 *	exists) that this channel no longer needs to be informed if the
 *	interpreter is deleted.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	In the future, if the interpreter is deleted this channel will
 *	no longer be informed.
 *
 *----------------------------------------------------------------------
 */

static void
X25ServerCloseProc(
    ClientData callbackData)	/* The data passed in the call to
                                 * Tcl_CreateCloseHandler. */
{
    X25AcceptCallback *acceptCallbackPtr;
    				/* The actual data. */

    acceptCallbackPtr = (X25AcceptCallback *) callbackData;
    if (acceptCallbackPtr->interp != (Tcl_Interp *) NULL) {
        X25UnregisterServerInterpCleanupProc(acceptCallbackPtr->interp,
                acceptCallbackPtr);
    }
    Tcl_EventuallyFree((ClientData) acceptCallbackPtr->script, TCL_DYNAMIC);
    ckfree((char *) acceptCallbackPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * X25UnregisterServerInterpCleanupProc --
 *
 *	Unregister a previously registered accept callback record. The
 *	interp field of this record will no longer be set to NULL in
 *	the future when the interpreter is deleted.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Prevents the interp field of the accept callback record from
 *	being set to NULL in the future when the interpreter is deleted.
 *
 *----------------------------------------------------------------------
 */

static void
X25UnregisterServerInterpCleanupProc(
    Tcl_Interp *interp,		/* Interpreter in which the accept callback
                                 * record was registered. */
    X25AcceptCallback *acceptCallbackPtr)
    				/* The record for which to delete the
                                 * registration. */
{
    Tcl_HashTable *hTblPtr;
    Tcl_HashEntry *hPtr;

    hTblPtr = (Tcl_HashTable *) Tcl_GetAssocData(interp,
            "x25AcceptCallbacks", NULL);
    if (hTblPtr == (Tcl_HashTable *) NULL) {
        return;
    }
    hPtr = Tcl_FindHashEntry(hTblPtr, (char *) acceptCallbackPtr);
    if (hPtr == (Tcl_HashEntry *) NULL) {
        return;
    }
    Tcl_DeleteHashEntry(hPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * X25BlockModeProc --
 *
 *	This procedure is invoked by the generic IO level to set blocking
 *	and nonblocking mode on an X.25 socket based channel.
 *
 * Results:
 *	0 if successful, errno when failed.
 *
 * Side effects:
 *	Sets the device into blocking or nonblocking mode.
 *
 *----------------------------------------------------------------------
 */

static int
X25BlockModeProc(instanceData, mode)
    ClientData instanceData;		/* Socket state. */
    int mode;				/* The mode to set. Can be one of
                                         * TCL_MODE_BLOCKING or
                                         * TCL_MODE_NONBLOCKING. */
{
    X25State *statePtr = (X25State *) instanceData;
    int setting;
    
    setting = fcntl(statePtr->fd, F_GETFL);
    if (mode == TCL_MODE_BLOCKING) {
        statePtr->flags &= (~(X25_ASYNC_SOCKET));
        setting &= (~(O_NONBLOCK));
    } else {
        statePtr->flags |= X25_ASYNC_SOCKET;
        setting |= O_NONBLOCK;
    }
    if (fcntl(statePtr->fd, F_SETFL, setting) < 0) {
        return errno;
    }

    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * X25WaitForConnect --
 *
 *	Waits for a connection on an asynchronously opened socket to
 *	be completed.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The socket is connected after this function returns.
 *
 *----------------------------------------------------------------------
 */

static int
X25WaitForConnect(statePtr, errorCodePtr)
    X25State *statePtr;		/* State of the socket. */
    int *errorCodePtr;		/* Where to store errors? */
{
    int timeOut;		/* How long to wait. */
    int state;			/* Of calling TclWaitForFile. */
    int flags;			/* fcntl flags for the socket. */

    /*
     * If an asynchronous connect is in progress, attempt to wait for it
     * to complete before reading.
     */
    
    if (statePtr->flags & X25_ASYNC_CONNECT) {
        if (statePtr->flags & X25_ASYNC_SOCKET) {
            timeOut = 0;
        } else {
            timeOut = -1;
        }
        errno = 0;
        state = TclUnixWaitForFile(statePtr->fd,
		TCL_WRITABLE | TCL_EXCEPTION, timeOut);
        if (!(statePtr->flags & X25_ASYNC_SOCKET)) {
            flags = fcntl(statePtr->fd, F_GETFL);
            flags &= (~(O_NONBLOCK));
            (void) fcntl(statePtr->fd, F_SETFL, flags);
        }
        if (state & TCL_EXCEPTION) {
            return -1;
        }
        if (state & TCL_WRITABLE) {
            statePtr->flags &= (~(X25_ASYNC_CONNECT));
        } else if (timeOut == 0) {
            *errorCodePtr = errno = EWOULDBLOCK;
            return -1;
        }
    }
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * X25InputProc --
 *
 *	This procedure is invoked by the generic IO level to read input
 *	from an X.25 socket based channel.
 *
 *	NOTE: We cannot share code with FilePipeInputProc because here
 *	we must use recv to obtain the input from the channel, not read.
 *
 * Results:
 *	The number of bytes read is returned or -1 on error. An output
 *	argument contains the POSIX error code on error, or zero if no
 *	error occurred.
 *
 * Side effects:
 *	Reads input from the input device of the channel.
 *
 *----------------------------------------------------------------------
 */

static int
X25InputProc(instanceData, buf, bufSize, errorCodePtr)
    ClientData instanceData;		/* Socket state. */
    char *buf;				/* Where to store data read. */
    int bufSize;			/* How much space is available
                                         * in the buffer? */
    int *errorCodePtr;			/* Where to store error code. */
{
    X25State *statePtr = (X25State *) instanceData;
    int bytesRead, state;

    *errorCodePtr = 0;
    state = X25WaitForConnect(statePtr, errorCodePtr);
    if (state != 0) {
        return -1;
    }
    bytesRead = recv(statePtr->fd, buf, bufSize, 0);
    if (bytesRead > -1) {
        return bytesRead;
    }
    if (errno == ECONNRESET) {

        /*
         * Turn ECONNRESET into a soft EOF condition.
         */
        return 0;
    }
    *errorCodePtr = errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * X25OutputProc --
 *
 *	This procedure is invoked by the generic IO level to write output
 *	to a X25 socket based channel.
 *
 *	NOTE: We cannot share code with FilePipeOutputProc because here
 *	we must use send, not write, to get reliable error reporting.
 *
 * Results:
 *	The number of bytes written is returned. An output argument is
 *	set to a POSIX error code if an error occurred, or zero.
 *
 * Side effects:
 *	Writes output on the output device of the channel.
 *
 *----------------------------------------------------------------------
 */

static int
X25OutputProc(instanceData, buf, toWrite, errorCodePtr)
    ClientData instanceData;		/* Socket state. */
    char *buf;				/* The data buffer. */
    int toWrite;			/* How many bytes to write? */
    int *errorCodePtr;			/* Where to store error code. */
{
    X25State *statePtr = (X25State *) instanceData;
    int written;
    int state;				/* Of waiting for connection. */

    *errorCodePtr = 0;
    state = X25WaitForConnect(statePtr, errorCodePtr);
    if (state != 0) {
        return -1;
    }
    written = send(statePtr->fd, buf, toWrite, 0);
    if (written > -1) {
        return written;
    }
    *errorCodePtr = errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * X25CloseProc --
 *
 *	This procedure is invoked by the generic IO level to perform
 *	channel-type-specific cleanup when a X25 socket based channel
 *	is closed.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the socket of the channel.
 *
 *----------------------------------------------------------------------
 */

static int
X25CloseProc(instanceData, interp)
    ClientData instanceData;	/* The socket to close. */
    Tcl_Interp *interp;		/* For error reporting - unused. */
{
    X25State *statePtr = (X25State *) instanceData;
    int errorCode = 0;

    /*
     * Delete a file handler that may be active for this socket if this
     * is a server socket - the file handler was created automatically
     * by Tcl as part of the mechanism to accept new client connections.
     * Channel handlers are already deleted in the generic IO channel
     * closing code that called this function, so we do not have to
     * delete them here.
     */
    
    Tcl_DeleteFileHandler(statePtr->fd);

    if (close(statePtr->fd) < 0) {
	errorCode = errno;
    }
    if( statePtr->numRemoteAddrs > 0 ) {
        ckfree(statePtr->remoteAddrList);
    }
    ckfree((char *) statePtr);

    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * X25GetOptionProc --
 *
 *	Computes an option value for a X25 socket based channel, or a
 *	list of all options and their values.
 *
 * Results:
 *	A standard Tcl result. The value of the specified option or a
 *	list of all options and	their values is returned in the
 *	supplied DString. Sets Error message if needed.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
X25GetOptionProc(instanceData, interp, optionName, dsPtr)
    ClientData instanceData;     /* Socket state. */
    Tcl_Interp *interp;          /* For error reporting - can be NULL. */
    char *optionName;	         /* Name of the option to
				  * retrieve the value for, or
				  * NULL to get all options and
				  * their values. */
    Tcl_DString *dsPtr;	         /* Where to store the computed
				  * value; initialized by caller. */
{
    X25State *statePtr = (X25State *) instanceData;
    CONN_ADR local;
    CONN_ADR remote;
    int link, lcn, foundOption;
    size_t len = 0;
    char buf[128];

    if (optionName != (char *) NULL) {
        len = strlen(optionName);
    }
    foundOption = 0;

    if ((len == 0) ||
            ((len > 1) && (optionName[1] == 'r') &&
                    (strncmp(optionName, "-remoteaddr", len) == 0))) {
        if( len == 0 ) {
            Tcl_DStringAppendElement(dsPtr, "-remoteaddr");
	}
	foundOption++;
	if( (ioctl(statePtr->fd, X25_RD_REMOTE_ADR, &remote)==0) &&
                (DecodeBCDAddr(buf,128,remote.host,remote.hostlen)==TCL_OK) ) {
	    Tcl_DStringAppendElement(dsPtr, buf);
	} else {
           if( interp ) {
	        Tcl_AppendResult(interp, "can't get remote X.25 address: ",
                        Tcl_PosixError(interp), (char *)NULL);
	    }
	    return TCL_ERROR;
	}
    }
    
    if ((len == 0) ||
            ((len > 1) && (optionName[1] == 'l') &&
                    (strncmp(optionName, "-localaddr", len) == 0))) {
        if( len == 0 ) {
            Tcl_DStringAppendElement(dsPtr, "-localaddr");
	}
	foundOption++;
	if( (ioctl(statePtr->fd, X25_RD_LOCAL_ADR, &local)==0) &&
                (DecodeBCDAddr(buf,128,local.host,local.hostlen)==TCL_OK) ) {
	    Tcl_DStringAppendElement(dsPtr, buf);
	} else {
            if( interp ) {
	        Tcl_AppendResult(interp, "can't get local X.25 address: ",
                        Tcl_PosixError(interp), (char *)NULL);
	    }
	    return TCL_ERROR;
	}
    }
    
    if ((len == 0) ||
            ((len > 1) && (optionName[1] == 'l') &&
                    (strncmp(optionName, "-link", len) == 0))) {
        if( len == 0 ) {
            Tcl_DStringAppendElement(dsPtr, "-link");
	}
	foundOption++;
	if( ioctl(statePtr->fd, X25_GET_LINK, &link)==0 ) {
	    sprintf(buf,"%d", link);
	    Tcl_DStringAppendElement(dsPtr, buf);
	} else {
            if( interp ) {
	        Tcl_AppendResult(interp, "can't get X.25 link number: ",
                        Tcl_PosixError(interp), (char *)NULL);
	    }
	    return TCL_ERROR;
	}
    }
    
    if ((len == 0) ||
            ((len > 1) && (optionName[1] == 'l') &&
                    (strncmp(optionName, "-lcn", len) == 0))) {
        if( len == 0 ) {
            Tcl_DStringAppendElement(dsPtr, "-lcn");
	}
	foundOption++;
	if( ioctl(statePtr->fd, X25_RD_LCGN, &lcn)==0 ) {
	    sprintf(buf,"%d", lcn);
	    Tcl_DStringAppendElement(dsPtr, buf);
	} else {
            if( interp ) {
	        Tcl_AppendResult(interp, "can't get logical channel number: ",
                        Tcl_PosixError(interp), (char *)NULL);
	    }
	    return TCL_ERROR;
	}
    }

    
    if( (len > 0) && (foundOption == 0) ) {
        return Tcl_BadChannelOption(interp, optionName,
                "remoteaddr localaddr link lcn");
    }

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * X25WatchProc --
 *
 *	Initialize the notifier to watch the fd from this channel.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Sets up the notifier so that a future event on the channel will
 *	be seen by Tcl.
 *
 *----------------------------------------------------------------------
 */

static void
X25WatchProc(instanceData, mask)
    ClientData instanceData;		/* The socket state. */
    int mask;				/* Events of interest; an OR-ed
                                         * combination of TCL_READABLE,
                                         * TCL_WRITABLE and TCL_EXCEPTION. */
{
    X25State *statePtr = (X25State *) instanceData;

    if (mask) {
	Tcl_CreateFileHandler(statePtr->fd, mask,
		(Tcl_FileProc *) Tcl_NotifyChannel,
		(ClientData) statePtr->channel);
    } else {
	Tcl_DeleteFileHandler(statePtr->fd);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * X25GetHandleProc --
 *
 *	Called from Tcl_GetChannelFile to retrieve OS handles from inside
 *	an X25 socket based channel.
 *
 * Results:
 *	Returns TCL_OK with the fd in handlePtr, or TCL_ERROR if
 *	there is no handle for the specified direction. 
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
X25GetHandleProc(instanceData, direction, handlePtr)
    ClientData instanceData;	/* The socket state. */
    int direction;		/* Not used. */
    ClientData *handlePtr;	/* Where to store the handle.  */
{
    X25State *statePtr = (X25State *) instanceData;

    *handlePtr = (ClientData)statePtr->fd;
fprintf(stderr,"DBG: X25GetHandleProc returning fd=%d\n",(int)*handlePtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DecodeBCDAddr --
 *
 *	Given a BCD-encoded host address, decode it to a null-delimited
 * 	ASCII string
 *
 * Results:
 *	Returns TCL_OK on success, TCL_ERROR otherwise.
 *
 * Side effects:
 *	addr string is set to the decoded address, null-delimited.
 *
 *----------------------------------------------------------------------
 */

static int
DecodeBCDAddr(char* addr,	/* buffer for decoded address with len >= 16 */
        int addrlen,		/* length of buffer for decoded address */
        const unsigned char* host, /* BCD-encoded address */
        int hostlen)		/* Length of BCD-encoded address, BCD digits */
{
    int i;
    unsigned char val;
    
    if( (hostlen < 1) || (addrlen < 16) ) {
        return TCL_ERROR;
    }

    for( i = 0; i < hostlen; i++ ) {
        if( i % 2 ) {
            val = (host[i >> 1] & 0xF);
	} else {
	    val = (host[i >> 1] >> 4);
	}
	addr[i] = (val > 9) ? ('a' + val - 10) : ('0' + val);
    }
    addr[i] = '\0';
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * EncodeBCDAddr --
 *
 *	Given an ASCII X.25 address, encode it into BCD digits and length.
 *
 * Results:
 *	Returns TCL_OK on success, TCL_ERROR otherwise.
 *
 * Side effects:
 *	dest buffer set to encoded address.  destlen set to number of BCD
 *	digits in the dest buffer.
 *
 *----------------------------------------------------------------------
 */

static int
EncodeBCDAddr(unsigned char *dest, /* buffer for BCD-encoded address */
        unsigned char *destlen, /* where to store length of BCD-coded address*/
        const char* addr)	/* null-terminated X.25 address to encoded */
{
    int i, nlen, addrlen;
    unsigned char val;
    
    *destlen = 0;
    if( addr == NULL ) {
        return TCL_ERROR;
    }
    addrlen = strlen(addr);
    if( (addrlen==0) || (addrlen > 15) ) {
        return TCL_ERROR;
    }
    for( nlen=0, i=0; i < addrlen; nlen++, i++ ) {
        val = addr[i];
	if( (val >= '0') && (val <= '9') ) {
	    val -= '0';
	} else if( (val >= 'A') && (val <= 'F') ) {
	    val = val - 'A' + 10;
	} else if( (val >= 'a') && (val <= 'f') ) {
	    val = val - 'a' + 10;
	} else {
	    *destlen = 0;
	    return TCL_ERROR;
	}
	if( i & 1 ) {
	    dest[i >> 1] |= val;
	} else {
	    dest[i >> 1] = (val << 4);
	}
    }
    *destlen = nlen;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CompareBCDAddr --
 *
 *	Given two BCD-encoded X.25 addresses, compare them.
 *
 * Results:
 *	Returns 0 if addresses are equal, non-zero if they differ.
 *
 * Side effects:
 *	none.
 *
 *----------------------------------------------------------------------
 */

static int
CompareBCDAddr(unsigned char len1,	/* length of first address 	*/
        unsigned char *addr1,		/* first address, BCD-encoded	*/
        unsigned char len2,		/* length of second address	*/
        unsigned char *addr2)		/* second address, BCD-encoded	*/
{
    int i;
    
    if( len1 != len2 ) {
        return 1;
    }
    if( (i = memcmp(addr1,addr2,(len1/2))) != 0 ) {
        return i;
    }
    if( (len1 % 2) != 0 ) {		/* odd number of BCD digits	*/
        i = len1/2;			/* ck upper 4 bits in last digit*/
        if( (addr1[i] & 0xF0) != (addr2[i] & 0xF0) ) {
	    return 1;
	}
    }
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * X25AddrListToBCDArray
 *
 *	This procedure converts a Tcl list of X.25 address strings into
 *	an array of BCD-encoded addresses in a CONN_ADR structure.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The variables pointed to by bcdArrPtr and bcdArrLenPtr are set
 *	when all elements in the list can be converted.  If *bcdArrPtr
 *	is non-null, its contents will be freed with ckfree() before
 *	being set to the new value.
 *
 *----------------------------------------------------------------------
 */

static int X25AddrListToBCDArray(
        Tcl_Interp *interp,		/* if non-NULL, used for error msgs */
	CONN_ADR *bcdArrPtr[],		/* where to store BCD-encoded array */
        int *bcdArrLenPtr,		/* where to store length of array */
	Tcl_Obj *addrList)		/* list of X.25 address strings	*/
{
    int i;
    int len = 0;
    Tcl_Obj *curAddr = NULL;
    CONN_ADR *bcdArr = NULL;

    if( *bcdArrPtr != NULL ) {		/* Clear out existing BCD Array	*/
        ckfree(*bcdArrPtr);
    }
    *bcdArrPtr = NULL;
    *bcdArrLenPtr = 0;
    
    if( addrList == NULL ) {		/* Check for NULL addrList ptr	*/
        return TCL_OK;
    }
    
    if( Tcl_ListObjLength(interp, addrList, &len) != TCL_OK ) {
        return TCL_ERROR;
    }
    if( len == 0 ) {			/* Check for Empty addrList	*/
        return TCL_OK;
    }

					/* Allocate new array and fill it*/
    if( (bcdArr = ckalloc(sizeof(CONN_ADR) * len)) == NULL ) {
        if( interp != NULL ) {
            Tcl_AppendResult(interp, "unable to allocate space for BCD array",
                    (char *) NULL);
	}
        return TCL_ERROR;
    }
    for( i = 0; i < len; i++ ) {
        Tcl_ListObjIndex(interp, addrList, i, &curAddr);
	if( EncodeBCDAddr(bcdArr[i].host, &(bcdArr[i].hostlen),
                Tcl_GetStringFromObj(curAddr,NULL)) != TCL_OK ) {
            if( interp != NULL ) {
	        Tcl_AppendResult(interp, "invalid X.25 address: ",
                        Tcl_GetStringFromObj(curAddr,NULL), (char *) NULL);
	    }
	    ckfree(bcdArr);
	    return TCL_ERROR;
	}
fprintf(stderr,"DBG: Added Addr to list: %s\n",Tcl_GetStringFromObj(curAddr,NULL));
    }
    *bcdArrLenPtr = len;
    *bcdArrPtr = bcdArr;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * X25BCDArrayToAddrList --
 *
 *	This procedure converts an array of BCD-encoded addresses in
 *	CONN_ADR structures to a Tcl list of X.25 address strings.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The Tcl_Obj list pointed to by addrListPtr is erased and set to
 *	a new list.
 *
 *---------------------------------------------------------------------- */

static int X25BCDArrayToAddrList(
        Tcl_Interp *interp,		/* if non-NULL, used for error msgs */
	Tcl_Obj *addrList,		/* dest for list of X.25 addr strings*/
	CONN_ADR bcdArr[],		/* BCD-encoded X.25 address array */
        int bcdArrLen)			/* length of abovel array */
{
    int i;
    char buf[20];

    Tcl_SetListObj(addrList, 0, NULL);	/* Clear out list object	*/
    if( (bcdArrLen > 0) && (bcdArr != NULL) ) {
        for( i = 0; i < bcdArrLen; i++ ) {
            if( DecodeBCDAddr(buf, 20, bcdArr[i].host, bcdArr[i].hostlen)
                    != TCL_OK ) {
                if( interp ) {
                    Tcl_AppendResult(interp, "unable to convert BCD address "
                            "to string", (char *) NULL);
		}
	        return TCL_ERROR;
	    }
	    if( Tcl_ListObjAppendElement(interp, addrList,
                    Tcl_NewStringObj(buf,NULL)) != TCL_OK ) {
	        return TCL_ERROR;
	    }
	}
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * X25CreateSocket --
 *
 *	Opens a X.25 client socket and returns a state structure around it
 *
 * Results:
 *	The state structure or NULL if failed.  An error message is returned
 *	in the interpreter on failure.
 *
 * Side effects:
 *	Opens a client socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

static X25State *
X25CreateSocket(Tcl_Interp *interp, /* For error reporting; can be NULL */
        Tcl_Obj *hostList,	/* List of remote X.25 addrs; can be NULL */
        int link,		/* Local X.25 link, or -1 for default */
        char *localaddr,	/* Local X.25 address; can be NULL */
        int server,		/* 0 for client socket, non-zero for server */
        int async)		/* 0 for synchronous client, non-zero async */
{
    int status, sock, asyncConnect, curState, origState, i;
    X25State *statePtr = NULL;
    CONN_DB bind_addr;			/* Socket address for in/out	*/
    CONN_ADR local_addr;		/* Local X.25 address		*/
    CONN_ADR *addrArray = NULL;		/* Array of remote X.25 Addresses */
    int numHosts = 0;			/* Length of addrArray */
    int pgrp = 0;			/* pid for setting process group */
    int user_apprv;			/* Call Accept done by user or X.25? */

					/* clear socket address	structs	*/
    memset( (char *)&bind_addr, 0, sizeof(CONN_DB) );
    memset( (char *)&local_addr, 0, sizeof(CONN_ADR) );

    sock = -1;
    origState = 0;
					/* Transfer ASCII list of addresses */
					/* into array of BCD-encoded addr's */
    if( hostList != NULL ) {
        if( X25AddrListToBCDArray(interp, &addrArray, &numHosts, hostList)
                != TCL_OK ) {
	    goto addressError;
	}
    }

    if( (localaddr != NULL) && EncodeBCDAddr(local_addr.host,
            &(local_addr.hostlen),localaddr) != TCL_OK ) {
	goto addressError;
    }
    
    if( (sock = socket(AF_X25, SOCK_STREAM, 0)) < 0 ) {
	goto addressError;
    }

    pgrp = getpid();			/* Set Process Group for socket */
    if( ioctl(sock, SIOCSPGRP, &pgrp) != 0 ) {
        goto addressError;
    }

    /*
     * Set the close-on-exec flag so that the socket will not get
     * inherited by child processes.
     */

    fcntl(sock, F_SETFD, FD_CLOEXEC);

#if 0
    /*
     * Set kernel space buffering
     */

    TclSockMinimumBuffers(sock, SOCKET_BUFSIZE);
#endif

    asyncConnect = 0;
    status = 0;
    if (server) {

	/*
	 * Set up to reuse server addresses automatically and bind to the
	 * specified port.
	 */
    
	status = 1;
        bind_addr.hostlen = ANY_SUBADDRESS | ANY_LINK;
#if 0
	(void) setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &status,
		sizeof(status));
#endif
	status = bind(sock, (struct sockaddr *) &bind_addr, sizeof(bind_addr));
	if( status != 0 ) {
            goto bindError;
	}
	if( link >= 0 ) {
            if( (status = ioctl(sock, X25_SET_LINK, (char *)&link)) != 0 ) {
	        goto bindError;
	    }
	}
	/*
	 * Set flag to indicate call approval packets should be sent by this
	 * process, not the base X.25 software.  See X25Accept below...
	 */
	user_apprv = 1;
	if( (status=ioctl(sock,X25_CALL_ACPT_APPROVAL,&user_apprv)) != 0 ) {
	    goto bindError;
	}

	/*
	 * Listen for new connections on our server socket
	 */
        status = listen(sock, 5);
fprintf(stderr,"DBG: X25CreateSocket listening at sock=%d st=%d\n",sock,status);  
	
    } else {
      
	if( numHosts != 1 ) {
            if( interp != NULL ) {
	        Tcl_AppendResult(interp,"client X.25 sockets must specify one "
                        " and only one remote address.", (char *) NULL);
	    }
	    status = -1;
	    goto bindError;
	}

	/*
	 * copy first host from list into the structure for our remote host
	 */
	bind_addr.hostlen = addrArray[0].hostlen;
	memcpy(&(bind_addr.host), &(addrArray[0].host), (MAXHOSTADR+1)/2);
	bind_addr.hostlen |= ANY_LINK;

	/*
	 * Deal with local address binding if necessary
	 */
	if( localaddr != NULL ) {
	    curState = 1;
	    (void) setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
                    (char *) &curState, sizeof(curState));
	    status = bind(sock, (struct sockaddr *) &local_addr,
		    sizeof(local_addr));
	    if (status < 0) {
		goto bindError;
	    }
	}

	/*
	 * If link was specified, set it
	 */
        if( link >= 0 ) {
            if( (status = ioctl(sock, X25_SET_LINK, (char *)&link)) != 0 ) {
                goto bindError;
            }
	}
	
	/*
	 * Attempt to connect. The connect may fail at present with an
	 * EINPROGRESS but at a later time it will complete. The caller
	 * will set up a file handler on the socket if she is interested in
	 * being informed when the connect completes.
	 */

        if (async) {
            origState = fcntl(sock, F_GETFL);
            curState = origState | O_NONBLOCK;
            status = fcntl(sock, F_SETFL, curState);
        } else {
            status = 0;
        }
        if (status > -1) {
            status = connect(sock, (struct sockaddr *) &bind_addr, 
                    sizeof(bind_addr));
            if (status < 0) {
                if (errno == EINPROGRESS) {
                    asyncConnect = 1;
                    status = 0;
                }
            }
        }
    }

bindError:
    if (status < 0) {
        if (interp != NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
	if( addrArray != NULL ) {
	    ckfree(addrArray);
	}
        if (sock != -1) {
            close(sock);
        }
        return NULL;
    }

    /*
     * Allocate a new X25State for this socket.
     */

    statePtr = (X25State *) ckalloc((unsigned) sizeof(X25State));
    statePtr->numRemoteAddrs = numHosts;
    statePtr->remoteAddrList = addrArray;
    statePtr->flags = 0;
    if( async ) {
        statePtr->flags |= X25_ASYNC_SOCKET;
    }
    if (asyncConnect) {
        statePtr->flags |= X25_ASYNC_CONNECT;
    }
    statePtr->fd = sock;
    
    return statePtr;

addressError:
    if( addrArray != NULL ) {
       ckfree(addrArray);
    }    
    if (sock != -1) {
        close(sock);
    }
    if (interp != NULL) {
	Tcl_AppendResult(interp, "couldn't open socket: ",
		Tcl_PosixError(interp), (char *) NULL);
    }
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * X25_OpenClient --
 *
 *	Opens a X.25 client socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed.  An error message is returned
 *	in the interpreter on failure.
 *
 * Side effects:
 *	Opens a client socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
X25_OpenClient(Tcl_Interp *interp,	/* For error reporting; can be NULL. */
        char *address,			/* Remote X.25 Address for call */
        int link,			/* Link for call or -1 for default */
        char *localaddr,		/* Local X.25 address for call */
        int async)			/* non-zero for asynchronous connect */
{
    X25State *statePtr;
    char channelName[20];
    Tcl_Obj *addrList, *addrStr;

    /*
     * Create a single-element list with the remote X.25 address
     */
    if( (addrStr = Tcl_NewStringObj(address,strlen(address))) == NULL ) {
        return NULL;
    }
    Tcl_IncrRefCount(addrStr);
    if( (addrList = Tcl_NewListObj(0,NULL)) == NULL ) {
        Tcl_DecrRefCount(addrStr);
        return NULL;
    }
    Tcl_IncrRefCount(addrList);
    if( Tcl_ListObjAppendElement(interp,addrList,addrStr) != TCL_OK ) {
        Tcl_DecrRefCount(addrList);
	Tcl_DecrRefCount(addrStr);
	return NULL;
    }

    /*
     * Create a new client socket and wrap it in a channel.
     */

    statePtr = X25CreateSocket(interp, addrList, link, localaddr, 0, async);
    Tcl_DecrRefCount(addrList);
    Tcl_DecrRefCount(addrStr);
    if (statePtr == NULL) {
	return NULL;
    }

    statePtr->acceptProc = NULL;
    statePtr->acceptProcData = (ClientData) NULL;

    sprintf(channelName, "x25s%d", statePtr->fd);

    statePtr->channel = Tcl_CreateChannel(&x25ChannelType, channelName,
            (ClientData) statePtr, (TCL_READABLE | TCL_WRITABLE));
    if (Tcl_SetChannelOption(interp, statePtr->channel, "-translation",
	    "auto crlf") == TCL_ERROR) {
        Tcl_Close((Tcl_Interp *) NULL, statePtr->channel);
        return NULL;
    }
    return statePtr->channel;
}

/*
 *----------------------------------------------------------------------
 *
 * X25_MakeClientChannel --
 *
 *	Creates a Tcl_Channel from an existing client X.25 socket.
 *
 * Results:
 *	The Tcl_Channel wrapped around the preexisting X.25 socket.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
X25_MakeClientChannel(ClientData sock) /* Socket fd to wrap as a channel */
{
    X25State *statePtr;
    char channelName[20];

    statePtr = (X25State *) ckalloc((unsigned) sizeof(X25State));
    statePtr->fd = (int) sock;
    statePtr->acceptProc = NULL;
    statePtr->acceptProcData = (ClientData) NULL;
    statePtr->numRemoteAddrs = 0;
    statePtr->remoteAddrList = NULL;

    sprintf(channelName, "x25s%d", statePtr->fd);
    
    statePtr->channel = Tcl_CreateChannel(&x25ChannelType, channelName,
            (ClientData) statePtr, (TCL_READABLE | TCL_WRITABLE));
    if (Tcl_SetChannelOption((Tcl_Interp *) NULL, statePtr->channel,
	    "-translation", "auto crlf") == TCL_ERROR) {
        Tcl_Close((Tcl_Interp *) NULL, statePtr->channel);
        return NULL;
    }
    return statePtr->channel;
}

/*
 *----------------------------------------------------------------------
 *
 * X25_OpenServer --
 *
 *	Opens a X.25 server socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed. If an error occurred, an
 *	error message is left in interp->result if interp is
 *	not NULL.
 *
 * Side effects:
 *	Opens a server socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
X25_OpenServer(Tcl_Interp *interp,	/* For error reporting - may be NULL */
        Tcl_Obj *hostList,		/* Remote X.25 address to listen for,
					 * or NULL if all */
        int link,			/* Link to listen on or -1 for all */
        char* localaddr,		/* Local address for listening */
        X25AcceptProc *acceptProc,	/* Callback for connection accept */
        ClientData acceptProcData)	/* Data for the callback routine */
{
    X25State *statePtr;
    char channelName[20];

    /*
     * Create a new server socket and wrap it in a channel.
     */

    statePtr = X25CreateSocket(interp, hostList, link, localaddr, 1, 0);
    if (statePtr == NULL) {
	return NULL;
    }

    statePtr->acceptProc = acceptProc;
    statePtr->acceptProcData = acceptProcData;

    /*
     * Set up the callback mechanism for accepting connections
     * from new clients.
     */
    
    Tcl_CreateFileHandler(statePtr->fd, TCL_READABLE, X25Accept,
            (ClientData) statePtr);
fprintf(stderr,"DBG: Created File Handler for X25Accept fd=%d\n",statePtr->fd);
    sprintf(channelName, "x25s%d", statePtr->fd);
    statePtr->channel = Tcl_CreateChannel(&x25ChannelType, channelName,
            (ClientData) statePtr, 0);
    return statePtr->channel;
}

/*
 *----------------------------------------------------------------------
 *
 * X25Accept --
 *	Accept an X.25 socket connection.  This is called by the event loop.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Creates a new connection socket. Calls the registered callback
 *	for the connection acceptance mechanism.
 *
 *----------------------------------------------------------------------
 */

static void
X25Accept(ClientData data,		/* Callback token, X25State ptr	*/
        int mask)			/* Not used			*/
{
    X25State *sockState;		/* Client data of server socket	*/
    int newsock = -1;			/* The new client socket	*/
    X25State *newSockState;		/* State for new socket.	*/
    CONN_DB  from;			/* Structure for accept() call	*/
    CONN_ADR local;			/* The local address		*/
    CONN_ADR remote;			/* The remote address		*/
    int len;				/* For accept interface		*/
    char remoteAddr[20];		/* decoded remote X.25 address	*/
    int link, lcn;			/* link/lcn used for virtual circuit */
    int addrMatch;			/* whether remote addr matched	*/
    char channelName[20];
    int accept_sent = 0;		/* whether Call Accept pkt sent */
    int i, j;

    sockState = (X25State *) data;

fprintf(stderr,"DBG: X25Accept start fd=%d\n",sockState->fd);

    len = sizeof(CONN_DB);
    if( (newsock=accept(sockState->fd, (struct sockaddr *)&from, &len)) < 0 ) {
        goto acceptError;
    }

fprintf(stderr,"DBG: X25Accept after accept() newsock=%d\n",newsock);

    /*
     * Ensure we have an accept procedure.  Then get link, logical
     * channel (lcn), and remote address from socket.
     */
    if( (sockState->acceptProc == NULL) ||
            (ioctl(newsock, X25_GET_LINK,      &link)   != 0) ||
            (ioctl(newsock, X25_RD_LCGN,       &lcn)    != 0) ||
            (ioctl(newsock, X25_RD_REMOTE_ADR, &remote) != 0) ) {
fprintf(stderr,"DBG: X25Accept failed first checks\n");
        goto acceptError;
    }

/* DBG */
fprintf(stderr,"DBG: remote host ");
for( j = 0; j < remote.hostlen; j += 2 ) {
    fprintf(stderr,"%02x",(remote.host)[j]);
}
fprintf(stderr,"\n");

    /*
     * Determine whether we matched any addresses from list of acceptable ones
     */
    if( sockState->numRemoteAddrs == 0 ) {
        addrMatch = 1;			/* everything matches		*/
    } else {
        addrMatch = 0;
        for( i = 0; i < sockState->numRemoteAddrs; i++ ) {
fprintf(stderr,"DBG: compare to [%d] ",i);
for( j = 0; j < sockState->remoteAddrList[i].hostlen; j+=2 ) {
    fprintf(stderr,"%02X",(sockState->remoteAddrList[i].host)[j]);
}
fprintf(stderr,"\n");
	    if( CompareBCDAddr(sockState->remoteAddrList[i].hostlen,
                    sockState->remoteAddrList[i].host, remote.hostlen,
                    remote.host) == 0 ) {
	        addrMatch = 1;
		break;
	    }
	}
    }
    if( !addrMatch ) {
fprintf(stderr,"DBG: X25Accept failed to find addrMatch\n");
        goto acceptError;
    }
fprintf(stderr,"DBG: X25Accept addrMatch succeeded\n");

    if( DecodeBCDAddr(remoteAddr, 20, remote.host, remote.hostlen)!=TCL_OK ) {
fprintf(stderr,"DBG: X25Accept DecodeBCDAddr failed\n");
        goto acceptError;
    }
    if( ioctl(newsock, X25_SEND_CALL_ACPT) != 0 ) {
fprintf(stderr,"DBG: X25Accept failed to send Call Acceptance packet\n");
        goto acceptError;
    }
    accept_sent = 1;
    
    /*
     * Set close-on-exec flag to prevent the newly accepted socket from
     * being inherited by child processes.
     */

    (void) fcntl(newsock, F_SETFD, FD_CLOEXEC);

    newSockState = (X25State *) ckalloc((unsigned) sizeof(X25State));

    newSockState->flags = 0;
    newSockState->fd = newsock;
    newSockState->acceptProc = (X25AcceptProc *) NULL;
    newSockState->acceptProcData = (ClientData) NULL;
    newSockState->numRemoteAddrs = 0;
    newSockState->remoteAddrList = NULL;

    sprintf(channelName, "x25s%d", newsock);
    newSockState->channel = Tcl_CreateChannel(&x25ChannelType, channelName,
	    (ClientData) newSockState, (TCL_READABLE | TCL_WRITABLE));

    Tcl_SetChannelOption((Tcl_Interp *) NULL, newSockState->channel,
	    "-translation", "auto crlf");

    if (sockState->acceptProc != (X25AcceptProc *) NULL) {
fprintf(stderr,"DBG: X25Accept calling acceptProc addr %s link %d lcn %d\n",
	remoteAddr, link, lcn);
	(sockState->acceptProc) (sockState->acceptProcData,
		newSockState->channel, remoteAddr, link, lcn);
    }
    return;

acceptError:
    if( newsock >= 0 ) {
        if( accept_sent != 0 ) {	/* Only write 0-len pkt after Call
					 * Accept has been sent */
            write(newsock,"",0);	/* required for correct socket close */
	}
	close(newsock);
    }
}

	/*--------------------------------------------------------------
	  x25.c
	  --------------------------------------------------------------*/
